#include <dlfcn.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "ap_config.h"
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_request.h"
#include "http_log.h"

#include <rsbac/types.h>
#include <rsbac/rc_data_structures.h>
#include <rsbac/getname.h>
#include <rsbac/rc_getname.h>
#include <rsbac/rc_types.h>
#include <rsbac/syscalls.h>
#include <rsbac/error.h>
#include <rsbac/cap_getname.h>
#include <rsbac/helpers.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "apr.h"
#include "apr_pools.h"

#define MODULE_SIGNATURE "mod_rsbac/0.1"
module AP_MODULE_DECLARE_DATA rsbac_module;

typedef struct {
	char *jail_dir;
	rsbac_jail_ip_t jail_rsbac_ip;
	rsbac_jail_flags_t jail_rsbac_flags;
	rsbac_cap_vector_t jail_rsbac_max_caps;
	rsbac_jail_scd_vector_t jail_rsbac_scd_get;
	rsbac_jail_scd_vector_t jail_rsbac_scd_modify;
	__u32 forced_srv_role;
} rsbac_srv_config;

typedef struct {
	apr_table_t  *forced_role;
	char *path;
} rsbac_dir_config;

typedef struct {
	int initcount;
} rsbac_ctx;

void str_rev(char * str)
{
	char c;
	char * end = str + strlen(str) - 1;
	for (;str < end; str++, end--) {
		c = *str;
		*str = *end;
		*end = c;
	}
}

void uint_to_str(unsigned int n, char * str)
{
	/*Quick and dirty way to do that, should be included elsewhere also */
	char * ptr = str;
	*ptr = '\0';
	if ( !n ) {
		*ptr++ = '0';
	} else {
		while(n) {
			*ptr++ = '0' + (n % 10);
			n /= 10;
		}
	}
	*ptr = '\0';
	str_rev(str);
}

int mod_rsbac_init_now(server_rec *s)
{
	apr_pool_t *pool=s->process->pool;
	rsbac_ctx *c;

	apr_pool_userdata_get((void **)&c, "rsbac_module_ctx", pool);
	if(c) {
		return (c->initcount++);
	}
	c = (rsbac_ctx *)apr_palloc(pool, sizeof(*c));
	c->initcount=1;
	apr_pool_userdata_set(c, "rsbac_module_ctx", apr_pool_cleanup_null,
			pool);
	return 0;
}

static int role_ct_srv(request_rec *r)
{
	rsbac_srv_config * srv = (rsbac_srv_config *) ap_get_module_config(r->server->module_config,
			&rsbac_module);
	int res = 0;
	//char *tmp ;
	rsbac_rc_role_id_t role = 0;
	rsbac_rc_role_id_t current;
	const char * tmp;

	role = (rsbac_rc_role_id_t)srv->forced_srv_role;
	res = rsbac_rc_get_current_role(&current);
	if (res) {
		ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
				"mod_rsbac: getting current role failed with error %i!",
				res);
		return DECLINED;
	}
	if (role > 0) {
		if ((role == current) && (res == 0)) {
			ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, 0, r->server,
				"mod_rsbac: role still %u, keeping the same role for this children",
				role);
			return DECLINED;
		}
#if (RSBAC_VERSION_MAJOR == 1) && (RSBAC_VERSION_MID == 2)
		res = rsbac_rc_change_role(role);
#else
		res = rsbac_rc_change_role(role, NULL);
#endif
		if (res) {
			ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
					"mod_rsbac: role change to %u denied by RC, incompatible role",
					role);
			return DECLINED;
		} else {
			ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, 0, r->server,
					"mod_rsbac: role change to %u granted by RC",
					role);
			return DECLINED;
		}
	} else {
		ap_log_error(APLOG_MARK, APLOG_ERR, 0, r->server,
				"mod_rsbac: no role change requested, current role is %u!",
				current);
	}
	return DECLINED;
}

static int role_ct_dir(request_rec *r)
{
	rsbac_dir_config * dir = (rsbac_dir_config *) ap_get_module_config(r->per_dir_config,
			&rsbac_module);
	int res = 0;
	//char *tmp ;
	rsbac_rc_role_id_t role = 0;
	rsbac_rc_role_id_t current;
	char *path;
	const char * tmp;

	// path = (char* )apr_palloc(r->pool, sizeof(r->filename)); 
	path = ap_make_dirstr_parent(r->pool, r->filename);    
	tmp = apr_table_get(dir->forced_role, path);

	/* per dir */
	if (tmp != NULL) {
		res = rsbac_rc_get_current_role(&current);
		role =  (rsbac_rc_role_id_t ) atol(apr_table_get(dir->forced_role, path));

		if ((role == current) && (res == 0)) {
			ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, r->server,
					"mod_rsbac: role still %u, keeping the same role for this children",
					role);
			return DECLINED;
		}

#if (RSBAC_VERSION_MAJOR == 1) && (RSBAC_VERSION_MID == 2)
		res = rsbac_rc_change_role(role);
#else
		res = rsbac_rc_change_role(role, NULL);
#endif

		if (res) {
			ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO ,0, r->server,	
					"mod_rsbac: role change to %u denied by RC, incompatible role while accessing %s",
					role, path);
			return DECLINED;
		} else {
			ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO ,0, r->server,	
					"mod_rsbac: role change to %u granted by RC",
					role);
			return DECLINED;
		}
	}
	return DECLINED;
}

static void *rsbac_create_srv_config(apr_pool_t *p, server_rec *s)
{
	rsbac_srv_config *cfg = apr_pcalloc(p, sizeof(rsbac_srv_config));
	cfg->jail_dir=NULL;
	/*We initialise some rsbac_jail args to defaults values */
	cfg->jail_rsbac_flags = 0;
#ifdef KERNEL_NEXT
	cfg->jail_rsbac_max_caps.cap[0] = (__u32) -1;
	cfg->jail_rsbac_max_caps.cap[1] = (__u32) -1;
#else
	cfg->jail_rsbac_max_caps = -1;
#endif
	cfg->jail_rsbac_scd_get = 0;
	cfg->jail_rsbac_scd_modify = 0;
	cfg->forced_srv_role = 0;
	return (void *)cfg;
}

static void *rsbac_create_dir_config(apr_pool_t *p, char *dir)
{
	rsbac_dir_config *cfg = (rsbac_dir_config *) apr_palloc(p, sizeof(rsbac_dir_config));
	cfg->forced_role = apr_table_make(p, 4);
	return (void *)cfg;
}

void *merge_rsbac_dir_configs (apr_pool_t *p, void *parent_dirv, void *subdirv)
{
	rsbac_dir_config *parent_dir = (rsbac_dir_config *)parent_dirv;
	rsbac_dir_config *subdir = (rsbac_dir_config *)subdirv;
	rsbac_dir_config *new =
		(rsbac_dir_config *)apr_palloc (p, sizeof(rsbac_dir_config));

	new->forced_role = apr_table_overlay (p, subdir->forced_role,
			parent_dir->forced_role);
	return (void *)new;
}

void *merge_rsbac_srv_configs (apr_pool_t *p, void *basev, void *overridesv)
{
	rsbac_srv_config *new = (rsbac_srv_config *)apr_pcalloc(p, sizeof(rsbac_srv_config));
	rsbac_srv_config *overrides = (rsbac_srv_config *) overridesv;

	new->forced_srv_role = overrides->forced_srv_role;

	return (void *)new;
}

static const char *cmd_jail_dir(cmd_parms *cmd, void *dummy, const char *p1)
{
	rsbac_srv_config *cfg = (rsbac_srv_config *)ap_get_module_config(cmd->server->module_config, &rsbac_module);
	const char *err;

	if((err=ap_check_cmd_context(cmd, GLOBAL_ONLY))) return err;
	if(!ap_is_directory(cmd->pool, p1)) {
		return "Jail to something that is not a directory";
	}
	cfg->jail_dir=(char *)p1;
	return NULL;
}

static const char *cmd_jail_addr(cmd_parms *cmd, void *dummy, const char *p1)
{
	rsbac_srv_config *cfg = (rsbac_srv_config *)ap_get_module_config(cmd->server->module_config, &rsbac_module);
	struct in_addr addr;
	const char *err;
	if((err=ap_check_cmd_context(cmd, GLOBAL_ONLY))) return err;
	if(!inet_aton(p1, &addr))
		return "Invalid IP address";
	cfg->jail_rsbac_ip = (unsigned long int)addr.s_addr;
	return NULL;
}
		
static const char *cmd_jail_flags(cmd_parms *cmd, void *dummy, const char *flag)
{
	rsbac_srv_config *cfg = (rsbac_srv_config *)ap_get_module_config(cmd->server->module_config, &rsbac_module);

	if(ap_strcmp_match(flag, "allow_external_ipc") == 0){
		cfg->jail_rsbac_flags |= JAIL_allow_external_ipc;
		return NULL;
	}
	if(ap_strcmp_match(flag,"allow_all_net_family") == 0){
		cfg->jail_rsbac_flags |= JAIL_allow_all_net_family;
		return NULL;
	}
	if(ap_strcmp_match(flag,"allow_inet_raw") == 0){
		cfg->jail_rsbac_flags |= JAIL_allow_inet_raw;
		return NULL;
	}
	if(ap_strcmp_match(flag,"allow_inet_localhost") == 0){
		cfg->jail_rsbac_flags |= JAIL_allow_inet_localhost;
		return NULL;
	}
	if(ap_strcmp_match(flag,"JAIL_allow_dev_read") == 0){
		cfg->jail_rsbac_flags |= JAIL_allow_dev_read;
		return NULL;
	}
	if(ap_strcmp_match(flag,"allow_dev_write") == 0){
		cfg->jail_rsbac_flags |= JAIL_allow_dev_write;
		return NULL;
	}
	if(ap_strcmp_match(flag,"allow_dev_get_status") == 0){
		cfg->jail_rsbac_flags |= JAIL_allow_dev_get_status;
		return NULL;
	}
	if(ap_strcmp_match(flag,"allow_dev_mod_system") == 0){
		cfg->jail_rsbac_flags |= JAIL_allow_dev_mod_system;
		return NULL;
	}
	if(ap_strcmp_match(flag,"allow_tty_open") == 0){
		cfg->jail_rsbac_flags |= JAIL_allow_tty_open;
		return NULL;
	}
	return "RsbacJailFlags: Invalid Flag";
}

static const char *
cmd_rc_role_assign_internal(cmd_parms *cmd, rsbac_dir_config *dir_cfg,
		const char *role, const int by_name)
{
	__u32 * role_array;
	__u32 role_number;
	union rsbac_rc_target_id_t tid;
	union rsbac_rc_item_value_t value;
	rsbac_list_ta_number_t ta_number;
	int i, nr_roles;
	char * path ;
	char  store[16]; /*ugly isn't it?*/ 

	ta_number =  0;
	role_number = atol(role);
	path =  ap_make_full_path(cmd->temp_pool,cmd->path,"");
	ap_no2slash(path);

	nr_roles = rsbac_rc_get_list(ta_number, RT_ROLE, &tid, RI_name, 0, NULL, NULL);
	role_array = (__u32 *) apr_palloc(cmd->temp_pool, nr_roles * sizeof(__u32));

	if (role_array == NULL)
		return "Out of Memory"; 

	nr_roles = rsbac_rc_get_list(ta_number, RT_ROLE,
			&tid,RI_name, nr_roles, role_array, NULL);

	for(i=0; i<nr_roles; i++) {
		tid.role = role_array[i];
		if(by_name) {
			if (rsbac_rc_get_item(ta_number, RT_ROLE,
						&tid, &tid, RI_name, &value, NULL) == 0){

				if(ap_strcmp_match(value.name, role) == 0) {
					uint_to_str(role_array[i],store);
					apr_table_set(dir_cfg->forced_role, path, store);
					return NULL; /*we assisgn the first name matched */
				}
			}
		} else {
			if ((ap_strcmp_match(role, "0") && !(role_number))) break;
			if(role_array[i] == role_number){
				uint_to_str(role_array[i], store);
				apr_table_set(dir_cfg->forced_role, path, store);
				return NULL;
			}
		}
	}

	return "Role supplied is invalid";
}

static const char *
cmd_rc_role_per_dir(cmd_parms *cmd, void *dummy,
		const char *role)
{
	rsbac_dir_config *dir_cfg = (rsbac_dir_config *)ap_get_module_config(cmd->context,
			&rsbac_module);
	rsbac_rc_role_id_t role_nr = 0;
	role_nr = atol(role);	

	if (role_nr == 0)
		return cmd_rc_role_assign_internal(cmd, dir_cfg, role, 1); 
	else
		return cmd_rc_role_assign_internal(cmd, dir_cfg, role, 0);   

	return "Argument must be Role or Number";
}

static const char *
cmd_rc_role_per_server(cmd_parms *cmd, void *dummy,
		const char *role)
{
	server_rec *s = cmd->server;
	rsbac_srv_config *srv_cfg = (rsbac_srv_config *)ap_get_module_config(s->module_config, &rsbac_module);
	rsbac_rc_role_id_t role_nr = 0;
	__u32 *role_array;
	union rsbac_rc_target_id_t tid;
	union rsbac_rc_item_value_t value;
	int i, nr_roles;

	role_nr = atol(role);
	nr_roles = rsbac_rc_get_list(0, RT_ROLE, &tid, RI_name, 0, NULL, NULL);
	role_array = (__u32 *) apr_palloc(cmd->temp_pool, nr_roles * sizeof(__u32));
	if (!role_array)
		return "Out of memory";
	nr_roles = rsbac_rc_get_list(0, RT_ROLE, &tid,RI_name, nr_roles, role_array, NULL);
	/* Converting name to role, first name matching is used */
	if (role_nr == 0)
		for(i=0; i<nr_roles; i++) {
			tid.role = role_array[i];
			if ((rsbac_rc_get_item(0, RT_ROLE,
				&tid, &tid, RI_name, &value, NULL) == 0) &&
		  		(ap_strcmp_match(value.name, role) == 0)) {
				srv_cfg->forced_srv_role = role_array[i];
				return NULL;
			}
		}
	else
	/* No conversion needed, just check if the role exists */
		for(i=0; i<nr_roles; i++) {
			tid.role = role_array[i];
			if (role_array[i] == role_nr) {
				srv_cfg->forced_srv_role = role_nr;
				return NULL;
			}
		}
	return "Role supplied is invalid or does not exist";
}

static int mod_rsbac_init(apr_pool_t *p, apr_pool_t *plog,
		apr_pool_t *ptemp, server_rec *s)
{
	rsbac_srv_config *cfg = (rsbac_srv_config *)ap_get_module_config(s->module_config, &rsbac_module);

	ap_add_version_component(p, MODULE_SIGNATURE);
	if(cfg->jail_dir == NULL) return OK;
	if(mod_rsbac_init_now(s) == 1) {
		if(rsbac_jail(RSBAC_JAIL_VERSION, cfg->jail_dir,
					cfg->jail_rsbac_ip, cfg->jail_rsbac_flags,
					cfg->jail_rsbac_max_caps, cfg->jail_rsbac_scd_get,
					cfg->jail_rsbac_scd_modify) !=0 ) {
			ap_log_error(APLOG_MARK, APLOG_ERR, 0, s, "could not jail to %s.",cfg->jail_dir);
			return !OK;
		}
		if(chdir("/") < 0) {
			ap_log_error(APLOG_MARK, APLOG_ERR, 0, s, "could not chdir to '/'.");
			return !OK;
		}
		ap_log_error(APLOG_MARK, APLOG_NOTICE|APLOG_NOERRNO, 0, s, "mod_rsbac: changed JAIL root to %s.",cfg->jail_dir);
		return OK;
	}
	else return OK;
}

static void register_hooks(apr_pool_t *p)
{
	/* we want to be loaded last, except for mod_fcgid... and other ? */
	static const char * const rsbacPost[] = {
		"mod_fcgid.c",
		"mod_cgid.c",
		NULL
	};

	/* mod_rsbac_init cares for jail */
	ap_hook_post_config(mod_rsbac_init, NULL, rsbacPost, APR_HOOK_REALLY_LAST);
	/* Care for RC roles */
	ap_hook_map_to_storage(role_ct_srv, NULL, NULL, APR_HOOK_LAST);
	ap_hook_header_parser(role_ct_dir, NULL, NULL, APR_HOOK_FIRST);
}

static const command_rec rsbac_cmds[] = {
	AP_INIT_TAKE1("JailDir", cmd_jail_dir, NULL, 
			RSRC_CONF, "path to a new root directory"),
	AP_INIT_TAKE1("RsbacJailAddr", cmd_jail_addr, NULL, 
			RSRC_CONF, "limit to IP address"),
	AP_INIT_ITERATE("RsbacJailFlags", cmd_jail_flags, NULL, 
			RSRC_CONF, "set restriction flags"),
	AP_INIT_TAKE1( "RcRole", cmd_rc_role_per_dir, NULL, 
			ACCESS_CONF,  "assign role [Name|Number] per directory"),
	AP_INIT_TAKE1( "PerServerRcRole", cmd_rc_role_per_server, NULL, 
			RSRC_CONF,  "assign role [Name|Number] per server"),
	{ NULL }
};

module AP_MODULE_DECLARE_DATA rsbac_module = {
	STANDARD20_MODULE_STUFF,
	rsbac_create_dir_config, /* create per-dir    config structures */
	merge_rsbac_dir_configs, /* merge  per-dir    config structures */
	rsbac_create_srv_config, /* create per-server config structures */
	merge_rsbac_srv_configs, /* merge  per-server config structures */
	rsbac_cmds,              /* table of config file commands       */
	register_hooks           /* register hooks                      */
};
